/******************************************************************************
*******************************************************************************
**
**  Copyright (C) 2006 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <syslog.h>
#include <pthread.h>

#include "gulm_defines.h"

int log_running = 0;
static unsigned int log_size;
static char *log;
char *syslog_msg;
unsigned int syslog_max_size;
static pthread_t log_thr;
static pthread_mutex_t log_wake_lock;
static pthread_cond_t log_wake_cond;
static char *head, *tail, *last;

/*
  The log is a circular buffer to hold syslog messages. The messages
  are stored as with a numeric char holding the priority followed by the
  message will a NULLs seperating the messages. Messages are allowed
  to wrap around the end.  The last byte in the buffer will always be
  NULL. This makes printing wraparound messages easier.  The log
  is never allowed to get so full that head wraps all the way around
  and is equal to tail.  There will always be at least a one byte space.
  This is so that you do not need another value to determine whether the log
  is totally full or totally empty.

  Message 
  -------
  Hddddddddddddddd0

  H: Message header byte (see below)
  d: Variable length text data, up to max_size - 1 in length
  0: NULL
  
  
  Message header byte (bits)
  --------------------------
  tlwuhppp

  t: truncated flag. set if the message has been truncated
  l: lost messages flag. set if a lost message counter follows.
  w: wrap flag. set if the message wraps around the end of the log.
  u: unused flag
  h: header flag. This is set on all headers. Its job is to keep header
     bytes from ever being equal to 0, so that I can use 0s to pad the log
     without them being confused for headers.
  p: priority bit.


  detail of tail end of log
  -------------------------
  ffffffrrHut

  f: Variable length free log space
  r: restricted log space.  messages can finish in the space, but no
     new messages are allowed to start here. (see explanation below)
  H: lost message header byte.
  u: unused space.  This is the location of the header when messages have been
     lost.
  f: tail pointer


  The restricted space is necessary becauase a log message needs at least
  three bytes.  So if the head ptr is within 4 bytes of the tail, it is
  impossible to write a message and still have the space necessary for the
  lost message header.  If a message comes it when the head ptr is within
  4 bytes of the tail, a couple of things can happen.

  If the head is at (tail - 1), then there must be a lost message header at
  (head - 1). There is nothing more to do.
  If the head is at (tail - 2), you write out a lost message header and
  increment the head to (tail - 1)
  if the head is at (tail - 3) or (tail - 4), you pad with NULLs until
  (tail - 2) and then you write out a lost message header, and increment head to
  (tail - 1)

*/
#define DEFAULT_MAX_MSG_SIZE 1024
#define DEFAULT_LOG_SIZE 8192

#define TRUNCATED_FLAG 0x80
#define LOST_MSGS_FLAG 0x40
#define WRAPPED_FLAG   0x20
#define HEADER_FLAG    0x08

#define LOST_MSGS_PRIO LOG_WARNING

#define log_empty (head == tail)

#define get_priority(x) ((int)(0x07 & *(unsigned char *)(x)))

#define truncated_flag(x) (TRUNCATED_FLAG & *(x))

#define lost_msgs_flag(x) (LOST_MSGS_FLAG & *(x))

#define wrapped_flag(x) (WRAPPED_FLAG & *(x))

#define inc(x) increment_ptr(&(x), 1)


/* This is written so that head and tail never point to an invalid
   address. There is no need to lock access to head or tail, as long as
   1. They never try to look or move backwards. 2. They always change from
   valid value to valid value. */
static void increment_ptr(char **ptr, unsigned int size)
{
	if (*ptr + size >= last)
		*ptr += (size - log_size);
	else
		*ptr += size;
}

static int get_free_space()
{
	int dist = tail - head;
	if (dist <= 0)
		dist += log_size;
	return dist;
}

/* This is called after the caller writes the message int the syslog_msg
   buffer. */
void log_to_buffer(int priority) 
{
	char header = (char)(priority & 0x07) | HEADER_FLAG;
	int free_space = get_free_space();
	syslog_msg[syslog_max_size] = 0; /* just to be paranoid */
	int len = strlen(syslog_msg) + 2; /* text + header + NULL */

	if (free_space == 1)
		return;

	/* If there's not enough room to write a message, write a
	   lost message header */
	if (free_space <= 4){
		while(free_space > 2){
			*head = 0;
			inc(head);
			free_space--;
		}
		header = LOST_MSGS_FLAG | HEADER_FLAG | LOST_MSGS_PRIO;
		*head = header;
		inc(head);
		goto wake_up;
	}

	if (len > free_space - 2){
		header |= TRUNCATED_FLAG;
		syslog_msg[free_space - 4] = 0;
		len = free_space - 2;
	}

	if (len > last - head){
		char *ptr = &syslog_msg[last - (head + 1)];
		header |= WRAPPED_FLAG;
		*head = header;
		memcpy(head + 1, syslog_msg, last - (head + 1));
		len -= last - head;
		memcpy(log, ptr, len);
		head = log + len;
	}
	else {
		*head = header;
		memcpy(head + 1, syslog_msg, len - 1);
		increment_ptr(&head, len);
	}
wake_up:
	pthread_mutex_lock(&log_wake_lock);
	pthread_cond_signal(&log_wake_cond);
	pthread_mutex_unlock(&log_wake_lock);
}


static void process_message(void)
{
	if (*tail == 0){
		inc(tail);
		return;
	}
	if (lost_msgs_flag(tail)){
		syslog(get_priority(tail), "**Log buffer out of space. Messages lost**\n");
		inc(tail);
		return;
	}
	syslog(get_priority(tail), "%s%s%s", tail + 1, wrapped_flag(tail)? log : "", truncated_flag(tail)? " ... **Log buffer out of space. Message truncated**\n" : "");
	if (wrapped_flag(tail))
		tail = strchr(log, 0) + 1;
	else {
		char *tmp = strchr(tail, 0) + 1;
		if (tmp == last)
			tail = log;
		else
			tail = strchr(tail, 0) + 1;
	}
}

static void *log_thread(void *unused)
{
	while (1) {
		pthread_mutex_lock(&log_wake_lock);
		if (log_empty)
			pthread_cond_wait(&log_wake_cond, &log_wake_lock);
		pthread_mutex_unlock(&log_wake_lock);

		while (!log_empty)
			process_message();
	}
	return (void *) 0;
}

void log_init(unsigned int my_log_size, unsigned int my_max_size)
{
	pthread_attr_t attr;	

	log_size = (my_log_size)? my_log_size : DEFAULT_LOG_SIZE;
	syslog_max_size = (my_max_size)? my_max_size : DEFAULT_MAX_MSG_SIZE;

	log = malloc(log_size);
	if (log == NULL)
		die(ExitGulm_InitFailed, "Failed to allocate memory for log ring buffer: %s\n", strerror(errno));
	syslog_msg = malloc(syslog_max_size);
	if (syslog_msg == NULL)
		die(ExitGulm_InitFailed, "Failed to allocate memory for log messages: %s\n", strerror(errno));
	head = tail = log;
	syslog_max_size--; /* You don't use the last byte of the msg. It's
			      always set to NULL to make printing easier */
	log_size--; /* You don't use the last byte of the log. It's always set
		       to NULL to make printing easier */
	*(syslog_msg + syslog_max_size) = 0;
	last = log + log_size;
	*last = 0;
	
	if (pthread_attr_init(&attr) != 0)
		die(ExitGulm_InitFailed, "Failed to initialize logging thread attribute structure: %s\n", strerror(errno));
	if (pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED) != 0)
		die(ExitGulm_InitFailed, "Failed to set logging thread detachstate attribute: %s\n", strerror(errno));
	pthread_mutex_init(&log_wake_lock, NULL);
	pthread_cond_init(&log_wake_cond, NULL);

	if (pthread_create(&log_thr, &attr, log_thread, NULL) != 0)
		die(ExitGulm_InitFailed, "Failed to start logging thread: %s\n", strerror(errno));
	log_running = 1;
}
